/*  Boolector: Satisfiablity Modulo Theories (SMT) solver.
 *
 *  Copyright (C) 2007-2009 Robert Daniel Brummayer.
 *  Copyright (C) 2007-2012 Armin Biere.
 *  Copyright (C) 2013-2016 Aina Niemetz.
 *  Copyright (C) 2012-2016 Mathias Preiner.
 *
 *  All rights reserved.
 *
 *  This file is part of Boolector.
 *  See COPYING for more information on using this software.
 */

#ifndef BTOR_PTR_HASH_H_INCLUDED
#define BTOR_PTR_HASH_H_INCLUDED

#include "utils/btormem.h"
#include "utils/btorhash.h"
#include "utils/btorhashptr.h"
#include <string.h>
#include <stdbool.h>
#include <assert.h>

/*------------------------------------------------------------------------*/

typedef struct BtorPtrHashTable BtorPtrHashTable;
typedef struct BtorPtrHashBucket BtorPtrHashBucket;

typedef void * (*BtorCloneKeyPtr) (BtorMemMgr *mm,
                                   const void * map, 
                                   const void * key);
typedef void (*BtorCloneDataPtr) (BtorMemMgr *mm, 
                                  const void * map, 
                                  BtorHashTableData * data,
                                  BtorHashTableData * cloned_data);

struct BtorPtrHashBucket
{
  /* public:
   */
  void *key;

  BtorHashTableData data;

  BtorPtrHashBucket *next;      /* chronologically */
  BtorPtrHashBucket *prev;      /* chronologically */

  /* private:
   */
  BtorPtrHashBucket *chain;     /* collision chain */
};

struct BtorPtrHashTable
{
  BtorMemMgr *mm;

  uint32_t size;
  uint32_t count;
  BtorPtrHashBucket **table;

  BtorHashPtr hash;
  BtorCmpPtr cmp;

  BtorPtrHashBucket *first;     /* chronologically */
  BtorPtrHashBucket *last;      /* chronologically */
};

/*------------------------------------------------------------------------*/

BtorPtrHashTable *btor_new_ptr_hash_table (BtorMemMgr *,
                                           BtorHashPtr, BtorCmpPtr);

/* Clone hash table. 'ckey' is a function mapping key to cloned key,
 * 'cdata' is a function mapping data to cloned data (note: as_ptr vs.
 * as_int!). 'key_map' represents a map mapping key to cloned key values.
 * 'data_map' represents a map mapping data to cloned data values. */
BtorPtrHashTable *btor_clone_ptr_hash_table (BtorMemMgr * mm, 
                                             BtorPtrHashTable * table, 
                                             BtorCloneKeyPtr ckey, 
                                             BtorCloneDataPtr cdata,
                                             const void * key_map, 
                                             const void * data_map);

void btor_delete_ptr_hash_table (BtorPtrHashTable * p2iht);

BtorPtrHashBucket * btor_get_ptr_hash_table (BtorPtrHashTable * p2iht,
                                             void * key);

BtorPtrHashBucket * btor_add_ptr_hash_table (BtorPtrHashTable * p2iht,
                                             void * key);

/* Remove from hash table the bucket with the key.  The key has to be an
 * element of the hash table.  If 'stored_data_ptr' is non zero, then data
 * to which the given key was mapped is copied to this location.   The same
 * applies to 'stored_key_ptr'.  If you traverse/iterate a hash table
 * through the chronological chains, then you can remove elements while
 * traversing the hash table.
 */
void btor_remove_ptr_hash_table (BtorPtrHashTable *,
                                 void * key,
                                 void ** stored_key_ptr,
                                 BtorHashTableData * stored_data_ptr);

unsigned btor_hash_str (const void * str);

#define btor_compare_str ((BtorCmpPtr)strcmp)

/*------------------------------------------------------------------------*/
/* iterators                                                               */
/*------------------------------------------------------------------------*/

#define BTOR_PTR_HASH_TABLE_ITERATOR_STACK_SIZE 8

typedef struct BtorPtrHashTableIterator
{
  BtorPtrHashBucket *bucket;
  void *cur;
  bool reversed;
  uint8_t num_queued;
  uint8_t pos;
  const BtorPtrHashTable *stack[BTOR_PTR_HASH_TABLE_ITERATOR_STACK_SIZE];
} BtorPtrHashTableIterator;

void btor_init_ptr_hash_table_iterator (BtorPtrHashTableIterator * it,
                                        const BtorPtrHashTable * t);
void btor_init_reversed_ptr_hash_table_iterator (BtorPtrHashTableIterator * it,
                                                 const BtorPtrHashTable * t);
void btor_queue_ptr_hash_table_iterator (BtorPtrHashTableIterator * it,
                                         const BtorPtrHashTable * t);
bool btor_has_next_ptr_hash_table_iterator (const BtorPtrHashTableIterator * it);
void * btor_next_ptr_hash_table_iterator (BtorPtrHashTableIterator * it);
BtorHashTableData * btor_next_data_ptr_hash_table_iterator (BtorPtrHashTableIterator * it);

/*------------------------------------------------------------------------*/
#endif
